@@ -3,9 +3,9 @@
 import asyncio
 import json
 import logging
-from asyncio import Condition, Lock
+from asyncio import Lock
 from enum import Enum
-from typing import Any, Dict, List, Union
+from typing import Any, Dict, List, Union, Optional
 
 import aiohttp
 from async_timeout import timeout
@@ -14,6 +14,9 @@ from .zone import Zone
 
 _LOG = logging.getLogger("pizone.controller")
 
+REFRESH_INTERVAL = 25.0
+
+
 class Controller:
     """Interface to IZone controller"""
     class Mode(Enum):
@@ -42,51 +45,74 @@ class Controller:
     """Cool-down period for retrying to connect to the controller"""
 
     _VALID_FAN_MODES = {
-        'disabled' : [Fan.LOW, Fan.MED, Fan.HIGH],
-        '3-speed' : [Fan.LOW, Fan.MED, Fan.HIGH, Fan.AUTO],
-        '2-speed' : [Fan.LOW, Fan.HIGH, Fan.AUTO],
-        'var-speed' : [Fan.LOW, Fan.MED, Fan.HIGH, Fan.AUTO],
-        } # type: Dict[str, List[Fan]]
+        'disabled': [Fan.LOW, Fan.MED, Fan.HIGH],
+        '3-speed': [Fan.LOW, Fan.MED, Fan.HIGH, Fan.AUTO],
+        '2-speed': [Fan.LOW, Fan.HIGH, Fan.AUTO],
+        'var-speed': [Fan.LOW, Fan.MED, Fan.HIGH, Fan.AUTO],
+        }  # type: Dict[str, List[Fan]]
 
-    def __init__(self, discovery, device_uid: str, device_ip: str) -> None:
-        """Create a controller interface. Usually this is called from the discovery service.
+    def __init__(self, discovery, device_uid: str,
+                 device_ip: str, is_v2: bool) -> None:
+        """Create a controller interface.
 
-        If neither device UID or address are specified, will search network
-        for exactly one controller. If UID is specified then the addr is ignored.
+        Usually this is called from the discovery service. If neither
+        device UID or address are specified, will search network for
+        exactly one controller. If UID is specified then the addr is
+        ignored.
 
         Args:
             device_uid: Controller UId as a string (eg: mine is '000013170')
                 If specified, will search the network for a matching device
-            device_addr: Device network address. Usually specified as IP address
+            device_addr: Device network address. Usually specified as IP
+                address
 
         Raises:
-            ConnectionAbortedError: If id is not set and more than one iZone instance is
-                discovered on the network.
-            ConnectionRefusedError: If no iZone discovered, or no iZone device discovered at
-                the given IP address or UId
+            ConnectionAbortedError: If id is not set and more than one iZone
+                instance is discovered on the network.
+            ConnectionRefusedError: If no iZone discovered, or no iZone
+                device discovered at the given IP address or UId
         """
         self._ip = device_ip
         self._discovery = discovery
         self._device_uid = device_uid
+        self._is_v2 = is_v2
 
-        self.zones = [] # type: List[Zone]
-        self.fan_modes = [] # type: List[Controller.Fan]
-        self._system_settings = {} # type: Controller.ControllerData
+        self.zones = []  # type: List[Zone]
+        self.fan_modes = []  # type: List[Controller.Fan]
+        self._system_settings = {}  # type: Controller.ControllerData
 
+        self._initialised = False
         self._fail_exception = None
-        self._reconnect_condition = Condition()
 
         self._sending_lock = Lock()
 
     async def _initialize(self) -> None:
-        """Initialize the controller, does not complete until the system is initialised."""
+        """Initialize the controller, does not complete until the system is
+        initialised."""
         await self._refresh_system(notify=False)
 
-        self.fan_modes = Controller._VALID_FAN_MODES[str(self._system_settings['FanAuto'])]
+        self.fan_modes = Controller._VALID_FAN_MODES[
+            str(self._system_settings['FanAuto'])]
 
         zone_count = int(self._system_settings['NoOfZones'])
         self.zones = [Zone(self, i) for i in range(zone_count)]
         await self._refresh_zones(notify=False)
+        self._initialised = True
+        if self._is_v2:
+            self.discovery.loop.create_task(self._poll_loop())
+
+    async def _poll_loop(self) -> None:
+        while True:
+            await asyncio.sleep(REFRESH_INTERVAL)
+
+            if self._discovery.is_closed:
+                return
+
+            try:
+                await self._refresh_all()
+                _LOG.debug("Polling unit %s.", self._device_uid)
+            except ConnectionError as ex:
+                _LOG.debug("Poll failed due to exeption %s.", repr(ex))
 
     @property
     def device_ip(self) -> str:
@@ -98,6 +124,11 @@ class Controller:
         '''UId of the unit'''
         return self._device_uid
 
+    @property
+    def is_v2(self) -> bool:
+        '''True if this is a v2 controller'''
+        return self._is_v2
+
     @property
     def discovery(self):
         return self._discovery
@@ -109,7 +140,8 @@ class Controller:
 
     async def set_on(self, value: bool) -> None:
         """Turn the system on or off."""
-        await self._set_system_state('SysOn', 'SystemON', 'on' if value else 'off')
+        await self._set_system_state(
+            'SysOn', 'SystemON', 'on' if value else 'off')
 
     @property
     def mode(self) -> 'Mode':
@@ -147,8 +179,9 @@ class Controller:
         """
         if value not in self.fan_modes:
             raise AttributeError("Fan mode {} not allowed".format(value.value))
-        await self._set_system_state('SysFan', 'SystemFAN', value.value,
-                                     'medium' if value is Controller.Fan.MED else value.value)
+        await self._set_system_state(
+            'SysFan', 'SystemFAN', value.value,
+            'medium' if value is Controller.Fan.MED else value.value)
 
     @property
     def sleep_timer(self) -> int:
@@ -165,7 +198,8 @@ class Controller:
         time = int(value)
         if time < 0 or time > 120 or time % 30 != 0:
             raise AttributeError(
-                'Invalid Sleep Timer \"{}\", must be divisible by 30'.format(value))
+                'Invalid Sleep Timer \"{}\", must be divisible by 30'
+                .format(value))
         await self._set_system_state('SleepTimer', 'SleepTimer', value, time)
 
     @property
@@ -183,47 +217,51 @@ class Controller:
         """Turn the free air system on or off.
         Async method, await to ensure command revieved by system.
         Raises:
-            AttributeError: If attempting to set the state of the free air system
-                when it is not enabled.
+            AttributeError: If attempting to set the state of the free air
+                system when it is not enabled.
         """
         if not self.free_air_enabled:
             raise AttributeError('Free air is disabled')
-        await self._set_system_state('FreeAir', 'FreeAir', 'on' if value else 'off')
+        await self._set_system_state(
+            'FreeAir', 'FreeAir', 'on' if value else 'off')
 
     @property
-    def temp_supply(self) -> float:
+    def temp_supply(self) -> Optional[float]:
         """Current supply, or in duct, air temperature."""
-        return float(self._get_system_state('Supply'))
+        return float(self._get_system_state('Supply')) or None
 
     @property
-    def temp_setpoint(self) -> float:
+    def temp_setpoint(self) -> Optional[float]:
         """AC unit setpoint temperature.
         This is the unit target temp with with rasMode == RAS,
         or with rasMode == master and ctrlZone == 13.
         """
-        return float(self._get_system_state('Setpoint'))
+        return float(self._get_system_state('Setpoint')) or None
 
     async def set_temp_setpoint(self, value: float):
         """AC unit setpoint temperature.
         This is the unit target temp with with rasMode == RAS,
         or with rasMode == master and ctrlZone == 13.
         Args:
-            value: Valid settings are between ecoMin and ecoMax, at 0.5 degree units.
+            value: Valid settings are between ecoMin and ecoMax, at
+            0.5 degree units.
         Raises:
             AttributeError: On setting if the argument value is not valid.
                 Can still be set even if the mode isn't appropriate.
         """
         if value % 0.5 != 0:
-            raise AttributeError('SetPoint \'{}\' not rounded to nearest 0.5'.format(value))
+            raise AttributeError(
+                'SetPoint \'{}\' not rounded to nearest 0.5'.format(value))
         if value < self.temp_min or value > self.temp_max:
-            raise AttributeError('SetPoint \'{}\' is out of range'.format(value))
-        await self._set_system_state('Setpoint', 'UnitSetpoint', value, str(value))
-
+            raise AttributeError(
+                'SetPoint \'{}\' is out of range'.format(value))
+        await self._set_system_state(
+            'Setpoint', 'UnitSetpoint', value, str(value))
 
     @property
-    def temp_return(self) -> float:
+    def temp_return(self) -> Optional[float]:
         """The return, or room, air temperature"""
-        return float(self._get_system_state('Temp'))
+        return float(self._get_system_state('Temp')) or None
 
     @property
     def eco_lock(self) -> bool:
@@ -233,21 +271,25 @@ class Controller:
     @property
     def temp_min(self) -> float:
         """The value for the eco lock minimum, or 15 if eco lock not set"""
-        return float(self._get_system_state('EcoMin')) if self.eco_lock else 15.0
+        return float(self._get_system_state('EcoMin')) \
+            if self.eco_lock else 15.0
 
     @property
     def temp_max(self) -> float:
         """The value for the eco lock maxium, or 30 if eco lock not set"""
-        return float(self._get_system_state('EcoMax')) if self.eco_lock else 30.0
+        return float(self._get_system_state('EcoMax')) \
+            if self.eco_lock else 30.0
 
     @property
     def ras_mode(self) -> str:
         """This indicates the current selection of the Return Air temperature Sensor.
         Possible values are:
-            master: the AC unit is controlled from a CTS, which is manually selected
+            master: the AC unit is controlled from a CTS, which is manually
+                selected.
             RAS:    the AC unit is controller from its own return air sensor
-            zones:  the AC unit is controlled from a CTS, which is automatically
-                selected dependant on the cooling/ heating need of zones.
+            zones:  the AC unit is controlled from a CTS, which is
+                automatically selected dependant on the cooling/ heating need
+                of zones.
         """
         return self._get_system_state('RAS')
 
@@ -264,22 +306,31 @@ class Controller:
 
     @property
     def zones_const(self) -> int:
-        """This indicates the number of constant zones the system is configured for."""
+        """This indicates the number of constant zones the system is
+        configured for."""
         return self._get_system_state('NoOfConst')
 
     @property
     def sys_type(self) -> str:
         """This indicates the type of the iZone system connected. Possible values are:
-        110: the system is zone control only and all the zones are OPEN/CLOSE zones
-        210: the system is zone control only. Zones can be temperature controlled,
-            dependant on the zone settings.
+        110: the system is zone control only and all the zones
+            are OPEN/CLOSE zones
+        210: the system is zone control only. Zones can be temperature
+            controlled, dependant on the zone settings.
         310: the system is zone control and unit control.
         """
         return self._get_system_state('SysType')
 
+    async def _refresh_all(self, notify: bool = True) -> None:
+        zones = int(self._system_settings['NoOfZones'])
+        await asyncio.gather(
+            self._refresh_system(notify),
+            *[self._refresh_zone_group(i, notify)
+              for i in range(0, zones, 4)])
+
     async def _refresh_system(self, notify: bool = True) -> None:
         """Refresh the system settings."""
-        values = await self._get_resource('SystemSettings') # type: Controller.ControllerData
+        values = await self._get_resource('SystemSettings')  # type: Controller.ControllerData  # noqa: E501
         if self._device_uid != values['AirStreamDeviceUId']:
             _LOG.error("_refresh_system called with unmatching device ID")
             return
@@ -297,24 +348,19 @@ class Controller:
 
     async def _refresh_zone_group(self, group: int, notify: bool = True):
         assert group in [0, 4, 8]
-        zone_data_part = await self._get_resource("Zones{0}_{1}".format(group+1, group+4))
+        zone_data_part = await self._get_resource(
+            "Zones{0}_{1}".format(group+1, group+4))
 
         for i in range(min(len(self.zones)-group, 4)):
             zone_data = zone_data_part[i]
-            self.zones[i+group]._update_zone(zone_data, notify) #pylint: disable=protected-access
-
-    async def _trigger_reconnect(self):
-        async with self._reconnect_condition:
-            self._reconnect_condition.notify()
+            self.zones[i+group]._update_zone(zone_data, notify)  # pylint: disable=protected-access  # noqa: E501
 
     def _refresh_address(self, address):
         """Called from discovery to update the address"""
-        if self._ip == address:
-            return
         self._ip = address
         # Signal to the retry connection loop to have another go.
         if self._fail_exception:
-            self._discovery.create_task(self._trigger_reconnect())
+            self._discovery.create_task(self._retry_connection())
 
     def _get_system_state(self, state):
         self._ensure_connected()
@@ -328,7 +374,7 @@ class Controller:
 
         async with self._sending_lock:
             await self._send_command_async(command, send)
-            
+
             # Need to refresh immediately after setting.
             try:
                 await self._refresh_system()
@@ -337,81 +383,74 @@ class Controller:
 
     def _ensure_connected(self) -> None:
         if self._fail_exception:
-            raise ConnectionError("Unable to connect to the controller") from self._fail_exception
+            raise ConnectionError("Unable to connect to the controller") \
+                from self._fail_exception
 
     def _failed_connection(self, ex):
         if self._fail_exception:
             self._fail_exception = ex
             return
         self._fail_exception = ex
+        if not self._initialised:
+            return
         self._discovery.controller_disconnected(self, ex)
-        self._discovery.create_task(self._retry_connection())
 
     async def _retry_connection(self) -> None:
-        while True:
-            await self._discovery.rescan()
+        _LOG.info(
+            "Attempting to reconnect to server uid=%s ip=%s",
+            self.device_uid, self.device_ip)
 
-            try:
-                async with timeout(Controller.CONNECT_RETRY_TIMEOUT):
-                    # event will be fired if reconnected
-                    async with self._reconnect_condition:
-                        await self._reconnect_condition.wait()
-            except asyncio.TimeoutError:
-                pass
+        try:
+            await self._refresh_all(notify=False)
 
-            _LOG.info("Attempting to reconnect to server uid=%s ip=%s",
-                      self.device_uid, self.device_ip)
+            self._fail_exception = None
 
-            try:
-                await asyncio.gather(self._refresh_system(notify=False),
-                                     self._refresh_zones(notify=False))
-
-                self._fail_exception = None
-
-                self._discovery.controller_reconnected(self)
-                self._discovery.controller_update(self)
-                for zone in self.zones:
-                    self._discovery.zone_update(self, zone)
-                break
-            except ConnectionError as ex:
-                # Expected, just carry on.
-                _LOG.warning("Reconnect attempt for uid=%s failed with exception: %s",
-                             self.device_uid, ex.__repr__())
+            self._discovery.controller_update(self)
+            for zone in self.zones:
+                self._discovery.zone_update(self, zone)
+            self._discovery.controller_reconnected(self)
+        except ConnectionError as ex:
+            # Expected, just carry on.
+            _LOG.warning(
+                "Reconnect attempt for uid=%s failed with exception: %s",
+                self.device_uid, ex.__repr__())
 
     async def _get_resource(self, resource: str):
         try:
             session = self._discovery.session
-            async with session.get('http://%s/%s' % (self.device_ip, resource),
-                                   timeout=Controller.REQUEST_TIMEOUT) as response:
-                return await response.json()
+            async with session.get(
+                    'http://%s/%s' % (self.device_ip, resource),
+                    timeout=Controller.REQUEST_TIMEOUT) as response:
+                return await response.json(content_type=None)
         except (asyncio.TimeoutError, aiohttp.ClientError) as ex:
             self._failed_connection(ex)
-            raise ConnectionError("Unable to connect to the controller") from ex
+            raise ConnectionError("Unable to connect to the controller") \
+                from ex
 
     async def _send_command_async(self, command: str, data: Any):
-        body = {command : data}
-        _LOG.info("POST to host: %s command: %s", self.device_ip, json.dumps(body))
+        # For some reason aiohttp fragments post requests, which causes
+        # the server to fail disgracefully. Implimented rough and dirty
+        # HTTP POST client.
+        loop = self.discovery.loop
+        on_complete = loop.create_future()
+        device_ip = self.device_ip
 
-        # For some reason aiohttp fragments post requests, which causes the server
-        # to fail disgracefully. Implimented rough and dirty HTTP POST client.
         class _PostProtocol(asyncio.Protocol):
-            def __init__(self, host, target, data, on_complete):
-                body = json.dumps(data)
-                self.message = (
-                    "POST " + target + " HTTP/1.1\r\n" +
-                    "Host: " + host +"\r\n" +
-                    "Content-Length: " + str(len(body)) + "\r\n" +
-                    "\r\n" + body)
-
-                self.on_complete = on_complete
-
             def connection_made(self, transport):
-                transport.write(self.message.encode())
+                body = json.dumps({command: data}).encode()
+                header = (
+                    "POST /" + command + " HTTP/1.1\r\n" +
+                    "Host: " + device_ip + "\r\n" +
+                    "Content-Length: " + str(len(body)) + "\r\n" +
+                    "\r\n").encode()
+                _LOG.debug(
+                    "Writing message to " + device_ip + body.decode())
+                transport.write(header + body)
                 self.transport = transport
 
             def data_received(self, data):
                 self.transport.close()
-                response = data.decode() # type: str
+                response = data.decode()  # type: str
                 lines = response.split('\r\n', 1)
                 if not lines:
                     return
@@ -419,22 +458,19 @@ class Controller:
                 if(len(parts) != 3):
                     return
                 if int(parts[1]) != 200:
-                    self.on_complete.set_exception(
-                        aiohttp.ClientResponseError(None, None, 
-                                                    status=int(parts[1]),
-                                                    message=parts[2]))
+                    on_complete.set_exception(
+                        aiohttp.ClientResponseError(
+                            None, None,
+                            status=int(parts[1]),
+                            message=parts[2]))
                 else:
-                    self.on_complete.set_result(True)
+                    on_complete.set_result(True)
 
         try:
-            loop = asyncio.get_event_loop()
-            on_complete = loop.create_future()
-
             async with timeout(Controller.REQUEST_TIMEOUT) as cm:
-                transport, _ = await loop.create_connection(
-                    lambda: _PostProtocol(self.device_ip, '/' + command, 
-                                          body, on_complete), 
-                                          self.device_ip, 80)
+                transport, _ = await loop.create_connection(  # type: ignore
+                    lambda: _PostProtocol(),
+                    self.device_ip, 80)  # mypy: ignore
 
                 # wait for response to be recieved.
                 await on_complete
